/*
************************************************************************
*
*   ExPaintSurf.c - PaintSurface command
*
*   Copyright (c) 1994-96
*
*   ETH Zuerich
*   Institut fuer Molekularbiologie und Biophysik
*   ETH-Hoenggerberg
*   CH-8093 Zuerich
*
*   SPECTROSPIN AG
*   Industriestr. 26
*   CH-8117 Faellanden
*
*   All Rights Reserved
*
*   Date of last modification : 96/05/01
*   Pathname of SCCS file     : /sgiext/molmol/src/cmdprim/SCCS/s.ExPaintSurf.c
*   SCCS identification       : 1.5
*
************************************************************************
*/

#include <cmd_prim.h>

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <values.h>

#include <break.h>
#include <grid.h>
#include <arg.h>
#include <par_hand.h>
#include <data_hand.h>
#include <prim_hand.h>
#include <attr_struc.h>
#include <graph_draw.h>

#define PN_SURFACE_COLOR "SurfacColor"

typedef struct {
  PrimObjP primP;
  GRID grid;
  float maxRad;
  DhAtomP atomP;
  float minDist;
} AtomData;

typedef struct {
  float val;
  float r, g, b;
} PotRange;

typedef struct {
  PotRange *rangeA;
  int rangeNo;
} PotData;

static int SurfPaint = 0;
static float MaxDist = 1.4;

static void
prepareAtom(DhAtomP atomP, void *clientData)
{
  AtomData *dataP = clientData;
  float rad;
  Vec3 x;

  if (BreakCheck(10000))
    return;

  rad = DhAtomGetAttr(atomP)->radius;
  if (rad < 0.0)
    return;

  DhAtomGetCoord(atomP, x);
  GridPrepareAddEntry(dataP->grid, x);

  if (rad > dataP->maxRad)
    dataP->maxRad = rad;
}

static void
insertAtom(DhAtomP atomP, void *clientData)
{
  AtomData *dataP = clientData;
  float rad;
  Vec3 x;

  if (BreakCheck(10000))
    return;

  rad = DhAtomGetAttr(atomP)->radius;
  if (rad < 0.0)
    return;

  DhAtomGetCoord(atomP, x);
  GridInsertEntry(dataP->grid, x, atomP);
}


static BOOL
getClosest(void *entryP, Vec3 x, void *clientData)
{
  DhAtomP atomP = entryP;
  AtomData *dataP = clientData;
  Vec3 dx;
  float rad, dist;

  DhAtomGetCoord(atomP, dx);
  Vec3Sub(dx, x);
  rad = DhAtomGetAttr(atomP)->radius;
  dist = dx[0] * dx[0] + dx[1] * dx[1] + dx[2] * dx[2] - rad * rad;

  if (dist < dataP->minDist) {
    dataP->atomP = atomP;
    dataP->minDist = dist;
  }

  return TRUE;
}

static void
setColorAtom(Vec3 col, Vec3 x, float pot, void *clientData)
{
  AtomData *dataP = clientData;
  AttrP attrP;

  if (BreakCheck(10000))
    return;

  dataP->atomP = NULL;
  dataP->minDist = MAXFLOAT;
  GridFind(dataP->grid, x, 1, getClosest, dataP);

  if (dataP->atomP == NULL)
    /* this should normally not happen, but give it the default
       color just in case */
    attrP = PrimGetAttr(dataP->primP);
  else
    attrP = DhAtomGetAttr(dataP->atomP);

  col[0] = attrP->colR;
  col[1] = attrP->colG;
  col[2] = attrP->colB;
}

static void
paintSurfOne(PrimObjP primP, void *clientData)
{
  PrimSurfaceSetColor(primP, NULL, NULL);
}

static void
paintSurfAtom(PrimObjP primP, void *clientData)
{
  DhMolP molP;
  PropRefP refP;
  AtomData data;

  if (BreakInterrupted())
    return;

  molP = PrimGetMol(primP);

  refP = PropGetRef(PROP_SELECTED, FALSE);

  data.primP = primP;
  data.grid = GridNew();
  data.maxRad = 0.0;
  DhMolApplyAtom(refP, molP, prepareAtom, &data);
  if (data.maxRad > 0.0) {
    GridInsertInit(data.grid, 1.1 * data.maxRad + MaxDist);
    DhMolApplyAtom(refP, molP, insertAtom, &data);
    PrimSurfaceSetColor(primP, setColorAtom, &data);
  } else {
    /* must be initialized, otherwise GridDestroy crashes */
    GridInsertInit(data.grid, 10.0);
  }

  GridDestroy(data.grid);
}

static PotData *
getPotData(void)
{
  PotData *potP;
  char *s, buf[50];
  int pos, i;

  potP = malloc(sizeof(*potP));
  potP->rangeNo = 1;
  potP->rangeA = malloc(sizeof(*potP->rangeA));
  potP->rangeA[0].val = 0.0;
  potP->rangeA[0].r = 0.0;
  potP->rangeA[0].g = 0.0;
  potP->rangeA[0].b = 1.0;

  s = ParGetStrVal(PN_SURFACE_COLOR);
  pos = 0;
  while (*s != '\0') {
    for (i = 0; i < sizeof(buf) - 1; i++) {
      if (*s == '\0' || isspace(*s))
	break;
      buf[i] = *s;
      s++;
    }
    buf[i] = '\0';

    if (pos == 0)
      potP->rangeA[potP->rangeNo - 1].val = atof(buf);
    else if (pos == 1)
      potP->rangeA[potP->rangeNo - 1].r = atof(buf);
    else if (pos == 2)
      potP->rangeA[potP->rangeNo - 1].g = atof(buf);
    else
      potP->rangeA[potP->rangeNo - 1].b = atof(buf);
    
    pos++;

    while (isspace(*s))
      s++;
    
    if (*s == '\0')
      break;

    if (pos == 4) {
      potP->rangeNo++;
      potP->rangeA = realloc(potP->rangeA,
	  potP->rangeNo * sizeof(*potP->rangeA));
      pos = 0;
    }
  }

  return potP;
}

static void
freePotData(PotData *potP)
{
  free(potP->rangeA);
  free(potP);
}

static void
setColorPot(Vec3 col, Vec3 x, float pot, void *clientData)
{
  PotData *potP = clientData;
  int rangeI;
  float par;

  if (BreakCheck(100000))
    return;

  rangeI = -1;
  while (rangeI < potP->rangeNo - 1 && pot > potP->rangeA[rangeI + 1].val)
    rangeI++;
  
  if (rangeI == -1) {
    col[0] = potP->rangeA[0].r;
    col[1] = potP->rangeA[0].g;
    col[2] = potP->rangeA[0].b;
  } else if (rangeI == potP->rangeNo - 1) {
    col[0] = potP->rangeA[rangeI].r;
    col[1] = potP->rangeA[rangeI].g;
    col[2] = potP->rangeA[rangeI].b;
  } else {
    par = (pot - potP->rangeA[rangeI].val) /
	(potP->rangeA[rangeI + 1].val - potP->rangeA[rangeI].val);
    col[0] = (1.0 - par) * potP->rangeA[rangeI].r +
	par * potP->rangeA[rangeI + 1].r;
    col[1] = (1.0 - par) * potP->rangeA[rangeI].g +
	par * potP->rangeA[rangeI + 1].g;
    col[2] = (1.0 - par) * potP->rangeA[rangeI].b +
	par * potP->rangeA[rangeI + 1].b;
  }
}

static void
paintSurfPot(PrimObjP primP, void *clientData)
{
  if (BreakInterrupted())
    return;

  PrimSurfaceSetColor(primP, setColorPot, clientData);
}

#define ARG_NUM 3
#define PAINT_NUM 3

ErrCode
ExPaintSurface(char *cmd)
{
  ArgDescr arg[ARG_NUM];
  EnumEntryDescr enumEntry[PAINT_NUM];
  ErrCode errCode;
  PotData *potP;
  PropRefP refP;

  arg[0].type = AT_ENUM;
  arg[1].type = AT_DOUBLE;
  arg[2].type = AT_STR;

  ArgInit(arg, ARG_NUM);

  enumEntry[0].str = "one_color";
  enumEntry[0].onOff = FALSE;
  enumEntry[1].str = "atom";
  enumEntry[1].onOff = FALSE;
  enumEntry[2].str = "pot";
  enumEntry[2].onOff = FALSE;

  enumEntry[SurfPaint].onOff = TRUE;

  arg[0].prompt = "Paint";
  arg[0].u.enumD.entryP = enumEntry;
  arg[0].u.enumD.n = PAINT_NUM;
  arg[0].v.intVal = SurfPaint;

  arg[1].prompt = "Max. Dist";
  arg[1].v.doubleVal = MaxDist;

  arg[2].prompt = "Pot. Colors";
  if (ParDefined(PN_SURFACE_COLOR))
    DStrAssignStr(arg[2].v.strVal, ParGetStrVal(PN_SURFACE_COLOR));
  else
    DStrAssignStr(arg[2].v.strVal, "-0.5 1 0 0 0.0 1 1 1 0.5 0 0 1");

  errCode = ArgGet(arg, ARG_NUM);
  if (errCode != EC_OK) {
    ArgCleanup(arg, ARG_NUM);
    return errCode;
  }

  SurfPaint = arg[0].v.intVal;
  MaxDist = arg[1].v.doubleVal;
  ParSetStrVal(PN_SURFACE_COLOR, DStrToStr(arg[2].v.strVal));

  ArgCleanup(arg, ARG_NUM);

  BreakActivate(TRUE);

  refP = PropGetRef(PROP_SELECTED, FALSE);
  if (SurfPaint == 0) {
    PrimApply(PT_SURFACE, refP, paintSurfOne, NULL);
  } else if (SurfPaint == 1) {
    PrimApply(PT_SURFACE, refP, paintSurfAtom, NULL);
  } else {
    potP = getPotData();
    PrimApply(PT_SURFACE, refP, paintSurfPot, potP);
    freePotData(potP);
  }

  BreakActivate(FALSE);

  GraphMolChanged(PROP_SELECTED);
  GraphRedrawNeeded();

  return EC_OK;
}
